Imports System
Imports System.Drawing
Imports System.Windows.Forms
Imports System.IO

' WinForm program for MSDN Magazine article "Image Recognition with the MNIST Data Set"
' Save as MnistViewer.vb
' You must get, download, and unzip the two MNIST data files below
' To compile, create a new VS project, then add this .vb file to the project
' OR use the VB compiler directly: >csc.exe /target:winexe MnistViewer.vb

Namespace MnistViewer
    Friend NotInheritable Class Program
        Private Sub New()
        End Sub

        <STAThread>
        Shared Sub Main()
            Application.EnableVisualStyles()
            Application.SetCompatibleTextRenderingDefault(False)
            Application.Run(New Form1)
        End Sub
    End Class

    Partial Public Class Form1
        Inherits Form

        ' get these files from http://yann.lecun.com/exdb/mnist/
        ' they must be unzipped after downloading
        ' edit paths as necessary

        Private pixelFile As String = "C:\MnistViewer\train-images.idx3-ubyte"
        Private labelFile As String = "C:\MnistViewer\train-labels.idx1-ubyte"
        Private trainImages() As DigitImage = Nothing

        Private components As System.ComponentModel.IContainer = Nothing

        Protected Overloads Overrides Sub Dispose(ByVal disposing As Boolean)
            If disposing AndAlso (components IsNot Nothing) Then
                components.Dispose()
            End If
            MyBase.Dispose(disposing)
        End Sub

        Public Sub New()
            InitializeComponent()
            textBox1.Text = pixelFile
            textBox2.Text = labelFile
            comboBox1.SelectedItem = "6" ' Magnification
            Me.ActiveControl = button1
        End Sub

        ' ==== UI methods

        Private Sub button1_Click(ByVal sender As Object, ByVal e As EventArgs) Handles button1.Click ' Load data
            Me.pixelFile = textBox1.Text
            Me.labelFile = textBox2.Text
            Me.trainImages = LoadData(pixelFile, labelFile)
            listBox1.Items.Add("MNIST images loaded into memory")
        End Sub

        Private Sub button2_Click(ByVal sender As Object, ByVal e As EventArgs) Handles button2.Click ' Display 'next' image
            Dim nextIndex As Integer = Integer.Parse(textBox4.Text)
            Dim currImage As DigitImage = trainImages(nextIndex)

            Dim mag As Integer = Integer.Parse(comboBox1.SelectedItem.ToString) ' magnification
            Dim bitMap As Bitmap = MakeBitmap(currImage, mag)
            pictureBox1.Image = bitMap

            Dim pixelVals As String = PixelValues(currImage)
            textBox5.Text = pixelVals

            textBox3.Text = textBox4.Text ' update curr idx from old next idz
            textBox4.Text = (nextIndex + 1).ToString ' ++next index

            listBox1.Items.Add("Curr image index = " & textBox3.Text & " label = " & currImage.label)
        End Sub

        ' ==== Code Logic methods =======================================================================

        'public static int ReverseBytes(int v) // 32 bits = 4 bytes
        '{
        '  // bit-manipulation version
        '  return (v & 0x000000FF) << 24 | (v & 0x0000FF00) << 8 |
        '         (v & 0x00FF0000) >> 8 | ((int)(v & 0xFF000000)) >> 24;
        '}

        Public Shared Function ReverseBytes(ByVal v As Integer) As Integer
            Dim intAsBytes() As Byte = BitConverter.GetBytes(v)
            Array.Reverse(intAsBytes)
            Return BitConverter.ToInt32(intAsBytes, 0)
        End Function

        Public Shared Function IntToBinaryString(ByVal v As Integer) As String
            ' to pretty print an int as binary
            Dim s As String = Convert.ToString(v, 2) ' base 2 but without leading 0s
            Dim t As String = s.PadLeft(32, "0"c) ' add leading 0s
            Dim res As String = ""
            For i = 0 To t.Length - 1
                If i > 0 AndAlso i Mod 8 = 0 Then
                    res &= " " ' add a space every 8 chars
                End If
                res &= t.Chars(i)
            Next i
            Return res
        End Function

        Public Shared Function LoadData(ByVal pixelFile As String, ByVal labelFile As String) As DigitImage()
            ' Load MNIST training set of 60,000 images into memory
            ' remove static to access listBox1
            Dim numImages As Integer = 60000
            Dim result(numImages - 1) As DigitImage

            Dim pixels(27) As Byte
            For i = 0 To pixels.Length - 1
                pixels(i) = New Byte(27) {}
            Next i

            Dim ifsPixels As New FileStream(pixelFile, FileMode.Open)
            Dim ifsLabels As New FileStream(labelFile, FileMode.Open)

            Dim brImages As New BinaryReader(ifsPixels)
            Dim brLabels As New BinaryReader(ifsLabels)

            Dim magic1 As Integer = brImages.ReadInt32 ' stored as Big Endian
            magic1 = ReverseBytes(magic1) ' convert to Intel format

            Dim imageCount As Integer = brImages.ReadInt32
            imageCount = ReverseBytes(imageCount)

            Dim numRows As Integer = brImages.ReadInt32
            numRows = ReverseBytes(numRows)
            Dim numCols As Integer = brImages.ReadInt32
            numCols = ReverseBytes(numCols)

            Dim magic2 As Integer = brLabels.ReadInt32
            magic2 = ReverseBytes(magic2)

            Dim numLabels As Integer = brLabels.ReadInt32
            numLabels = ReverseBytes(numLabels)

            ' each image
            For di = 0 To numImages - 1
                For i = 0 To 27 ' get 28x28 pixel values
                    For j = 0 To 27
                        Dim b As Byte = brImages.ReadByte
                        pixels(i)(j) = b
                    Next j
                Next i

                Dim lbl As Byte = brLabels.ReadByte ' get the label
                Dim dImage As New DigitImage(28, 28, pixels, lbl)
                result(di) = dImage
            Next di ' each image

            ifsPixels.Close()
            brImages.Close()
            ifsLabels.Close()
            brLabels.Close()

            Return result
        End Function ' LoadData

        Public Shared Function MakeBitmap(ByVal dImage As DigitImage, ByVal mag As Integer) As Bitmap
            ' create a VB Bitmap suitable for display in a PictureBox control
            Dim width As Integer = dImage.width * mag
            Dim height As Integer = dImage.height * mag
            Dim result As New Bitmap(width, height)
            Dim gr As Graphics = Graphics.FromImage(result)
            For i = 0 To dImage.height - 1
                For j = 0 To dImage.width - 1
                    Dim pixelColor As Integer = 255 - dImage.pixels(i)(j) ' white background, black digits
                    'int pixelColor = dImage.pixelsij; // black background, white digits
                    Dim c As Color = Color.FromArgb(pixelColor, pixelColor, pixelColor) ' gray scale
                    'Color c = Color.FromArgb(pixelColor, 0, 0); // red scale
                    Dim sb As New SolidBrush(c)
                    gr.FillRectangle(sb, j * mag, i * mag, mag, mag) ' fills bitmap via Graphics object
                Next j
            Next i
            Return result
        End Function

        Public Shared Function PixelValues(ByVal dImage As DigitImage) As String
            ' create a string, with embedded newlines, suitable 
            ' for display in a multi-line TextBox control
            Dim s As String = ""
            For i = 0 To dImage.height - 1
                For j = 0 To dImage.width - 1
                    s &= dImage.pixels(i)(j).ToString("X2") & " "
                Next j
                s &= Environment.NewLine
            Next i
            Return s
        End Function

        ' ===================

        Private textBox1 As System.Windows.Forms.TextBox
        Private textBox2 As System.Windows.Forms.TextBox
        Private WithEvents button1 As System.Windows.Forms.Button
        Private textBox3 As System.Windows.Forms.TextBox
        Private textBox4 As System.Windows.Forms.TextBox
        Private label1 As System.Windows.Forms.Label
        Private label2 As System.Windows.Forms.Label
        Private comboBox1 As System.Windows.Forms.ComboBox
        Private label3 As System.Windows.Forms.Label
        Private listBox1 As System.Windows.Forms.ListBox
        Private pictureBox1 As System.Windows.Forms.PictureBox
        Private WithEvents button2 As System.Windows.Forms.Button
        Private textBox5 As System.Windows.Forms.TextBox

        Private Sub InitializeComponent()
            Me.textBox1 = New System.Windows.Forms.TextBox
            Me.textBox2 = New System.Windows.Forms.TextBox
            Me.button1 = New System.Windows.Forms.Button
            Me.textBox3 = New System.Windows.Forms.TextBox
            Me.textBox4 = New System.Windows.Forms.TextBox
            Me.label1 = New System.Windows.Forms.Label
            Me.label2 = New System.Windows.Forms.Label
            Me.comboBox1 = New System.Windows.Forms.ComboBox
            Me.label3 = New System.Windows.Forms.Label
            Me.listBox1 = New System.Windows.Forms.ListBox
            Me.pictureBox1 = New System.Windows.Forms.PictureBox
            Me.button2 = New System.Windows.Forms.Button
            Me.textBox5 = New System.Windows.Forms.TextBox
            CType(Me.pictureBox1, System.ComponentModel.ISupportInitialize).BeginInit()
            Me.SuspendLayout()
            ' 
            ' textBox1
            ' 
            Me.textBox1.Location = New System.Drawing.Point(12, 22)
            Me.textBox1.Name = "textBox1"
            Me.textBox1.Size = New System.Drawing.Size(308, 20)
            Me.textBox1.TabIndex = 0
            Me.textBox1.Text = "C:\MnistViewer\train-images.idx3-ubyte"
            ' 
            ' textBox2
            ' 
            Me.textBox2.Location = New System.Drawing.Point(12, 48)
            Me.textBox2.Name = "textBox2"
            Me.textBox2.Size = New System.Drawing.Size(308, 20)
            Me.textBox2.TabIndex = 1
            Me.textBox2.Text = "C:\MnistViewer\train-labels.idx1-ubyte"
            ' 
            ' button1
            ' 
            Me.button1.Location = New System.Drawing.Point(326, 32)
            Me.button1.Name = "button1"
            Me.button1.Size = New System.Drawing.Size(93, 23)
            Me.button1.TabIndex = 2
            Me.button1.Text = "Load Images"
            Me.button1.UseVisualStyleBackColor = True
            '	  Me.button1.Click += New System.EventHandler(Me.button1_Click)
            ' 
            ' textBox3
            ' 
            Me.textBox3.Enabled = False
            Me.textBox3.Location = New System.Drawing.Point(12, 91)
            Me.textBox3.Name = "textBox3"
            Me.textBox3.Size = New System.Drawing.Size(62, 20)
            Me.textBox3.TabIndex = 3
            Me.textBox3.Text = "NA"
            ' 
            ' textBox4
            ' 
            Me.textBox4.Location = New System.Drawing.Point(80, 91)
            Me.textBox4.Name = "textBox4"
            Me.textBox4.Size = New System.Drawing.Size(62, 20)
            Me.textBox4.TabIndex = 4
            Me.textBox4.Text = "0"
            ' 
            ' label1
            ' 
            Me.label1.AutoSize = True
            Me.label1.Location = New System.Drawing.Point(9, 114)
            Me.label1.Name = "label1"
            Me.label1.Size = New System.Drawing.Size(55, 13)
            Me.label1.TabIndex = 5
            Me.label1.Text = "Curr Index"
            ' 
            ' label2
            ' 
            Me.label2.AutoSize = True
            Me.label2.Location = New System.Drawing.Point(77, 114)
            Me.label2.Name = "label2"
            Me.label2.Size = New System.Drawing.Size(58, 13)
            Me.label2.TabIndex = 6
            Me.label2.Text = "Next Index"
            ' 
            ' comboBox1
            ' 
            Me.comboBox1.FormattingEnabled = True
            Me.comboBox1.Items.AddRange(New Object() {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10"})
            Me.comboBox1.Location = New System.Drawing.Point(159, 91)
            Me.comboBox1.Name = "comboBox1"
            Me.comboBox1.Size = New System.Drawing.Size(62, 21)
            Me.comboBox1.TabIndex = 7
            ' 
            ' label3
            ' 
            Me.label3.AutoSize = True
            Me.label3.Location = New System.Drawing.Point(156, 115)
            Me.label3.Name = "label3"
            Me.label3.Size = New System.Drawing.Size(70, 13)
            Me.label3.TabIndex = 8
            Me.label3.Text = "Magnification"
            ' 
            ' listBox1
            ' 
            Me.listBox1.Font = New System.Drawing.Font("Courier New", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, (CByte(0)))
            Me.listBox1.FormattingEnabled = True
            Me.listBox1.HorizontalScrollbar = True
            Me.listBox1.ItemHeight = 14
            Me.listBox1.Location = New System.Drawing.Point(12, 443)
            Me.listBox1.Name = "listBox1"
            Me.listBox1.Size = New System.Drawing.Size(280, 102)
            Me.listBox1.TabIndex = 9
            ' 
            ' pictureBox1
            ' 
            Me.pictureBox1.BackColor = System.Drawing.SystemColors.ControlDark
            Me.pictureBox1.Location = New System.Drawing.Point(12, 143)
            Me.pictureBox1.Name = "pictureBox1"
            Me.pictureBox1.Size = New System.Drawing.Size(280, 280)
            Me.pictureBox1.TabIndex = 10
            Me.pictureBox1.TabStop = False
            ' 
            ' button2
            ' 
            Me.button2.Location = New System.Drawing.Point(245, 91)
            Me.button2.Name = "button2"
            Me.button2.Size = New System.Drawing.Size(75, 23)
            Me.button2.TabIndex = 11
            Me.button2.Text = "Display Next"
            Me.button2.UseVisualStyleBackColor = True
            '	  Me.button2.Click += New System.EventHandler(Me.button2_Click)
            ' 
            ' textBox5
            ' 
            Me.textBox5.Font = New System.Drawing.Font("Courier New", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, (CByte(0)))
            Me.textBox5.Location = New System.Drawing.Point(307, 143)
            Me.textBox5.Multiline = True
            Me.textBox5.Name = "textBox5"
            Me.textBox5.Size = New System.Drawing.Size(606, 412)
            Me.textBox5.TabIndex = 12
            ' 
            ' Form1
            ' 
            Me.AutoScaleDimensions = New System.Drawing.SizeF(6.0F, 13.0F)
            Me.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font
            Me.ClientSize = New System.Drawing.Size(929, 570)
            Me.Controls.Add(Me.textBox5)
            Me.Controls.Add(Me.button2)
            Me.Controls.Add(Me.pictureBox1)
            Me.Controls.Add(Me.listBox1)
            Me.Controls.Add(Me.label3)
            Me.Controls.Add(Me.comboBox1)
            Me.Controls.Add(Me.label2)
            Me.Controls.Add(Me.label1)
            Me.Controls.Add(Me.textBox4)
            Me.Controls.Add(Me.textBox3)
            Me.Controls.Add(Me.button1)
            Me.Controls.Add(Me.textBox2)
            Me.Controls.Add(Me.textBox1)
            Me.Name = "Form1"
            Me.Text = "MNIST Viewer"
            CType(Me.pictureBox1, System.ComponentModel.ISupportInitialize).EndInit()
            Me.ResumeLayout(False)
            Me.PerformLayout()
        End Sub

        ' ===================
    End Class ' class Form1

    Public Class DigitImage
        ' an MNIST image of a '0' thru '9' digit
        Dim width As Integer ' 28
        Dim height As Integer ' 28
        Dim pixels(,) As Byte ' 0(white) - 255(black)
        Dim label As Byte ' '0' - '9'

        Public Sub New(ByVal width As Integer, ByVal height As Integer, ByVal pixels(,) As Byte, ByVal label As Byte)
            Me.width = width
            Me.height = height
            Me.pixels = New Byte(height - 1)() {}
            For i = 0 To Me.pixels.Length - 1
                Me.pixels(i) = New Byte(width - 1) {}
            Next i

            For i = 0 To height - 1
                For j = 0 To width - 1
                    Me.pixels(i)(j) = pixels(i)(j)
                Next j
            Next i

            Me.label = label
        End Sub
    End Class
End Namespace ' ns
